from .. import Bolors
from web3 import Web3
from web3.exceptions import TransactionNotFound

from .flashbots import TxParams


class TxBuilder:
    def __init__(self, flashEnabled_w3: Web3):
        """
        :param flashEnabled_w3: flashbot enabled Web3 module
        """
        if "flashbots" not in flashEnabled_w3:
            print(f"{Bolors.FAIL}Error there is no flashbot interface loaded. {Bolors.RESET}")
            exit(3)

        self._w3 = flashEnabled_w3
        self._sender = flashEnabled_w3.eth.account.address
        self.nonce = flashEnabled_w3.eth.get_transaction_count(self._sender)
        self.gas = 0
        self.gasPrice = 0
        self.gasMaxPrice = 0
        self.chainID = 1
        self.type = 2
        self.bundle = []

    def OverrideGasConfig(self, gas: int, gas_price: int, max_price: int) -> "TxBuilder":
        """
        the override the configuration for the gas amount and the gas price
        :param gas:
        :param gas_price:
        :param max_price:
        :return:
        """
        self.gas = gas
        self.gasPrice = gas_price
        self.gasMaxPrice = max_price
        return self

    def setChainId(self, id: int) -> "TxBuilder":
        self.chainID = id
        return self

    def setType(self, n: int) -> "TxBuilder":
        self.type = n
        return self

    # bundle two EIP-1559 (type 2) transactions, pre-sign one of them
    # NOTE: chainId is necessary for all EIP-1559 txns
    # NOTE: nonce is required for signed txns

    def append(self, tx: TxParams) -> "TxBuilder":
        """
        bundle two EIP-1559 (type 2) transactions, pre-sign one of them
        NOTE: chainId is necessary for all EIP-1559 txns
        NOTE: nonce is required for signed txns

        :param tx: the standard transaction parameters
        :return:
        """
        tx["type"] = self.type
        tx["chainId"] = self.chainID
        tx["maxFeePerGas"] = self.gasMaxPrice
        tx["maxPriorityFeePerGas"] = self.gasPrice
        tx["gas"] = self.gas
        tx["nouce"] = self.nonce
        self.bundle.append(tx)
        self.nonce = self.nonce + 1
        return self

    def exe(self):
        # keep trying to send bundle until it gets mined
        while True:
            block = self._w3.eth.block_number
            print(f"Simulating on block {block}")
            # simulate bundle on current block
            try:
                self._w3.flashbots.simulate(self.bundle, block)
                print("Simulation successful.")
            except Exception as e:
                print("Simulation error", e)
                return

            # send bundle targeting next block
            print(f"Sending bundle targeting block {block + 1}")
            send_result = self._w3.flashbots.send_bundle(self.bundle, target_block_number=block + 1)
            send_result.wait()
            try:
                receipts = send_result.receipts()
                print(f"\nBundle was mined in block {receipts[0].blockNumber}\a")
                break
            except TransactionNotFound:
                print(f"Bundle not found in block {block + 1}")

        print(
            f"Sender account balance: {Web3.fromWei(self._w3.eth.get_balance(self._sender), 'ether')} ETH"
        )
        print(
            f"Receiver account balance: {Web3.fromWei(self._w3.eth.get_balance(self._sender), 'ether')} ETH"
        )



import codecs
import json
import os
import subprocess
import time
from threading import Lock
from typing import List, Tuple, Optional

import re

# ========================== Of course
from hexbytes import HexBytes
from web3 import Web3, HTTPProvider
from web3._utils.module import attach_modules
from web3.contract import Contract as Web3Contract
from web3.datastructures import AttributeDict
from web3.exceptions import TransactionNotFound, ContractLogicError, InvalidAddress, TimeExhausted
from web3.logs import DISCARD
from web3.middleware import geth_poa_middleware
from web3.types import BlockData

# ========================== Of course
from . import Bolors, Evm, DefaultKeys, root_base_path, MetaSetting
from .buildercompile.remotecompile import BuildRemoteLinuxCommand
from .buildercompile.transpile import BuildLang, filter_file_name, BuildLangForge
from .conf import Config
from .exceptions import FoundUndeployedLibraries
from .flashbots import construct_flashbots_middleware, flashbot
from .paths import Paths


def web3_provider(address: str) -> Web3:
    try:
        if address.startswith('http'):  # HTTP
            return Web3(Web3.HTTPProvider(address))
        if address.startswith('ws'):  # WebSocket
            return Web3(Web3.WebsocketProvider(address))
        return Web3(Web3.IPCProvider(address))
    except FileNotFoundError:
        raise ValueError("Failed to initialize web3 provider (is eth_node set?)") from None


w3_lock = Lock()
event_lock = Lock()
statement = 'End : {}, IO File {}'


def extract_tx_by_address(address, block: BlockData) -> list:
    # Note: block attribute dict has to be generated with full_transactions=True flag
    return [tx for tx in block.transactions if tx.to and address.lower() == tx.to.lower()]


def event_log(tx_hash: str, events: List[str], provider: Web3, contract: Web3Contract) -> Tuple[str, Optional[AttributeDict]]:
    """
    Extracts logs of @event from tx_hash if present
    :param tx_hash:
    :param events: Case sensitive events name
    :param provider:
    :param contract: Web3 Contract
    :return: event name and log represented in 'AttributeDict' or 'None' if not found
    """
    try:
        receipt = provider.eth.getTransactionReceipt(tx_hash)
    except TransactionNotFound:
        time.sleep(3000)  # hard coded sleep for 3 seconds... maybe this will help?
        # retry
        try:
            receipt = provider.eth.getTransactionReceipt(tx_hash)
        except TransactionNotFound:
            return '', None

    for event in events:
        # we discard warning as we do best effort to find wanted event, not always there
        # as we listen to the entire contract tx, might
        log = getattr(contract.events, event)().processReceipt(receipt, DISCARD)
        if log:
            data_index = 0
            return event, log[data_index]
    # todo: fix this - seems like some weird return
    return '', None


def normalize_address(address: str):
    """Converts address to address acceptable by web3"""
    return Web3.toChecksumAddress(address.lower())


def Logd(anystr: any):
    print(anystr)


def writeFile(content, filename):
    fo = open(filename, "w")
    fo.write(content)
    fo.close()
    print(statement.format(time.ctime(), filename))


class HexJsonEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, HexBytes):
            return obj.hex()
        return super().default(obj)


def _parseValue(val):
    # check for nested dict structures to iterate through
    if 'dict' in str(type(val)).lower():
        return toDict(val)
    # convert 'HexBytes' type to 'str'
    elif 'HexBytes' in str(type(val)):
        return val.hex()
    else:
        return val


def toDict(dictToParse):
    # convert any 'AttributeDict' type found to 'dict'
    parsedDict = dict(dictToParse)
    for key, val in parsedDict.items():
        if 'list' in str(type(val)):
            parsedDict[key] = [_parseValue(x) for x in val]
        else:
            parsedDict[key] = _parseValue(val)
    return parsedDict


# binops
regex1 = r"\/\/*.*"
# solidity version cutter
regex2 = r"^(\d+\.)?(\d+\.)?(\*|\d+)"


class IDos:
    def hasContractName(self, name: str) -> bool:
        pass

    def getAddr(self, keyname: str) -> str:
        pass

    def isAddress(self, add: str) -> bool:
        pass


class BinOp:
    """
    The binary operation for the ops. Taking care the operations for library linking and the related stuffs.
    """

    def __init__(self, bin_content: str, file_name: str):
        self.bin_raw = bin_content
        self.bin_knifed = bin_content
        self.bin_undeploy_lib = dict()
        self.file_name = file_name
        self.debug = False

    def setDebug(self, de: bool):
        self.debug = de

    def GetRawBin(self) -> str:
        return self.bin_raw

    def GetKnifedBin(self) -> str:
        return self.bin_knifed

    def checkBinForUndeployLib(self) -> bool:
        matches = re.finditer(regex1, self.bin_raw, re.MULTILINE)
        found = False
        for matchNum, match in enumerate(matches, start=1):
            print("Library {matchNum} is found at {start}-{end}: {match}".format(matchNum=matchNum, start=match.start(), end=match.end(), match=match.group()))
            k, v = self.fromLine(match.group())
            self.bin_undeploy_lib[k] = v
            found = True
        return found

    def fromLine(self, input_line: str) -> Tuple[str, str]:
        class_name = input_line.split(":")[1]
        return class_name, input_line

    def _placehd(self, instruction_line: str) -> str:
        return "__{}__".format(str(instruction_line).split("->")[0].strip(" //"))

    def anaylze(self, databank: IDos) -> bool:
        if len(self.bin_undeploy_lib) == 0:
            print("🚧 Nothing to process")
            return False

        for class_name, instruction_line in self.bin_undeploy_lib.items():
            if databank.hasContractName(class_name) is True:
                print(f"💽 Found support Class {class_name} - deployment address")
                if databank.isAddress(databank.getAddr(class_name)):
                    self._knifeBinClass(class_name, self._placehd(instruction_line), databank.getAddr(class_name))
                else:
                    print("🧊 The found library address is not valid - {}, {}".format(class_name, databank.getAddr(class_name)))
                    raise FoundUndeployedLibraries
            else:
                print("⚠️ Unfound library Error- {}, please make sure you have this library deployed.".format(class_name))
                raise FoundUndeployedLibraries

        self.bin_knifed = self.bin_knifed.splitlines(True)[0]
        self.bin_knifed = self.bin_knifed.replace("\n", "")
        # self.bin_knifed = "0x" + self.bin_knifed
        if self.debug is True:
            print(f"After processed bin file - {self.file_name}.bin (should be done now)")
            print(self.bin_knifed)
            # print(self.bin_raw)
            print("File content end ##")
        else:
            print(f"After processed bin file - {self.file_name}.bin")

        return True

    def _knifeBinClass(self, c: str, k: str, address: str) -> None:
        # address_step_1 = address.lower()
        address_step_2 = address.replace("0x", "")
        self.bin_knifed = self.bin_knifed.replace(k, address_step_2)
        print(f"🍡 Linked successfully for Solidity Class {c} with {address}")


class SolWeb3Tool(object):
    """
    This is the tool to build operation of the compiling solidity contract source code
    Try to make some improvement of code to make better access
    This is the artifact manager as we know it
    """
    OUTPUT_BUILD = "build"
    WORKSPACE_PATH = ""
    solfolder = ""
    file_name = "xxx.sol"
    prefixname = ""
    statement = 'End : {}, IO File {}'

    def __init__(self):
        self._abi = None
        self._bin = None
        self._meta = None
        self.combined_data = None
        self._key = None

    def setBuildNameSpace(self, path: str) -> "SolWeb3Tool":
        self.OUTPUT_BUILD = path
        return self

    def setBasePath(self, path: str) -> "SolWeb3Tool":
        self.WORKSPACE_PATH = path
        return self

    def SplitForgeBuild(self, class_name: str) -> "SolWeb3Tool":
        uncutjson = dict()
        combinedjson = os.path.join(self.WORKSPACE_PATH, self.OUTPUT_BUILD, "{}.sol".format(class_name), "{}.json".format(class_name))
        try:
            uncutjson = json.load(codecs.open(combinedjson, 'r', 'utf-8-sig'))
        except FileNotFoundError:
            print("Some of the files from the build in forge is not found")
            exit(3)
        abifile = os.path.join(self.WORKSPACE_PATH, self.OUTPUT_BUILD, "{}.sol".format(class_name), "{}.abi".format(class_name))
        binfile = os.path.join(self.WORKSPACE_PATH, self.OUTPUT_BUILD, "{}.sol".format(class_name), "{}.bin".format(class_name))

        if "abi" in uncutjson:
            predum = uncutjson["abi"]
            writeFile(json.dumps(predum, ensure_ascii=False), abifile)

        if "bytecode" in uncutjson:
            pr = uncutjson["bytecode"]
            if "object" in pr:
                pr2 = pr["object"]
                pr2 = pr2.replace("0x", "")
                writeFile(pr2, binfile)
            if "linkReferences" in pr:
                links = pr["linkReferences"]
                for a in links:
                    print("found link")

        return self

    def GetCodeClassFromForgeBuild(self, class_name: str) -> "SolWeb3Tool":
        """
        get the independent files and content from the file system
        :param class_name:
        :return:
        """
        p2abi = os.path.join(self.WORKSPACE_PATH, self.OUTPUT_BUILD, "{}.sol".format(class_name), "{}.abi".format(class_name))
        p1bin = os.path.join(self.WORKSPACE_PATH, self.OUTPUT_BUILD, "{}.sol".format(class_name), "{}.bin".format(class_name))
        try:
            self._bin = codecs.open(p1bin, 'r', 'utf-8-sig').read()
            self._abi = json.load(codecs.open(p2abi, 'r', 'utf-8-sig'))

        except FileNotFoundError:
            print("The artifacts files are not found from forge builds")
            exit(3)
        return self

    def GetCodeClassFromBuild(self, class_name: str) -> "SolWeb3Tool":
        """
        get the independent files and content from the file system
        :param class_name:
        :return:
        """
        p1bin = os.path.join(self.WORKSPACE_PATH, self.OUTPUT_BUILD, "{}.bin".format(class_name))
        p2abi = os.path.join(self.WORKSPACE_PATH, self.OUTPUT_BUILD, "{}.abi".format(class_name))
        metafile = os.path.join(self.WORKSPACE_PATH, self.OUTPUT_BUILD, "{}_meta.json".format(class_name))
        try:
            self._bin = codecs.open(p1bin, 'r', 'utf-8-sig').read()
            self._abi = json.load(codecs.open(p2abi, 'r', 'utf-8-sig'))
            self._meta = json.load(codecs.open(metafile, 'r', 'utf-8-sig'))
        except FileNotFoundError:
            print("The artifacts files are not found from solc builds")
            exit(3)
        return self

    def LoadInternalMeta(self, class_name: str) -> "SolWeb3Tool":
        metafile = os.path.join(root_base_path, self.OUTPUT_BUILD, "{}_meta.json".format(class_name))
        self._meta = json.load(codecs.open(metafile, 'r', 'utf-8-sig'))
        return self

    def GetMetadata(self) -> dict:
        return self._meta

    def GetSourceFileRead(self, file_name: str) -> str:
        asfile = os.path.join(self.WORKSPACE_PATH, file_name)
        return self.ReadAsStr(asfile)

    def ReadAsStr(self, file_name: str) -> str:
        return codecs.open(file_name, 'r', 'utf-8-sig').read()

    def ReadAsStrAndEscape(self, file_name: str) -> str:
        loaded = self.ReadAsStr(file_name)
        return re.escape(loaded)

    def GetMetaCompilerVer(self, full: bool = False) -> str:
        if "compiler" not in self._meta:
            print("key compiler is not found")
            return ""
        if "version" not in self._meta["compiler"]:
            print("key version is not found")
            return ""

        version_text = self._meta["compiler"]["version"]
        matches = re.search(regex2, version_text)

        if full is True:
            return version_text.replace('.Emscripten.clang', '')
        else:
            return matches.group()

    def GetMetaSettings(self) -> any:

        if "settings" not in self._meta:
            print("key settings is not found")
            return False
        if "evmVersion" not in self._meta["settings"]:
            print("key version is not found")
            return False
        if "libraries" not in self._meta["settings"]:
            print("key version is not found")
            return False

        if "optimizer" not in self._meta["settings"]:
            print("key version is not found")
            return False

        meta = MetaSetting(
            evm=self._meta["settings"]["evmVersion"],
            solidity_ver=self.GetMetaCompilerVer(False),
            solidity_ver_full=self.GetMetaCompilerVer(True),
            linkLib=self._meta["settings"]["libraries"],
            optimization_runs=self._meta["settings"]["optimizer"]["runs"],
            optimization_enabled=self._meta["settings"]["optimizer"]["enabled"]
        )

        return meta

    def GetCombinedFile(self) -> "SolWeb3Tool":
        pathc = os.path.join(self.WORKSPACE_PATH, self.OUTPUT_BUILD, "combined.json")
        try:
            pathcli = codecs.open(pathc, 'r', 'utf-8-sig')
            self.combined_data = json.load(pathcli)
        except Exception as e:
            print("Problems from loading items from the file: ", e)
        return self

    def byClassName(self, path: str, classname: str) -> str:
        # generating the string with path and class name
        return "{prefix}:{name}".format(prefix=path, name=classname)

    def GetCodeTag(self, fullname) -> [str, str]:
        """
        Search for the abi session and the bin session from the meta source file
        from combined.json
        :param fullname: initial file name
        :return: abi code and the bin code
        """
        return self.combined_data["contracts"][fullname]["abi"], self.combined_data["contracts"][fullname]["bin"]

    def GetCode(self, path: str, classname: str) -> [str, str]:
        """
        Search for the abi session and the bin session from the meta source file
        get the code and abi from combined.json
        :param path:
        :param classname:
        :return:
        """
        return self.GetCodeTag(self.byClassName(path, classname))

    def CompileBash(self) -> None:
        """
        This is the remote command to execute the solc_remote bash file
        using remote compile method to compile the sol files
        all works will be done with the remote server or using the docker
        """
        list_files = subprocess.run(["{}/solc_remote".format(self.WORKSPACE_PATH)])
        print("The exit code was: %d" % list_files.returncode)

    @property
    def abi(self) -> str:
        return self._abi

    @property
    def bin(self) -> str:
        return self._bin

    @property
    def workspace(self) -> str:
        return self.WORKSPACE_PATH

    def StoreTxResult(self, tx_result_data: any, filepath: str) -> None:
        """
        Having the result of the transaction data to be stored in an external JSON file.
        :param tx_result_data: input data
        :param filepath: the file name path
        :return: nothing to return
        """
        predump = toDict(tx_result_data)
        writeFile(json.dumps(predump, ensure_ascii=False), filepath)


class MiliDoS(IDos):
    """
    This is the base package function core for all the related operations to execute
    The center hub of the progress source code is in here
    """

    EVM_VERSION = Evm.BERLIN

    def __init__(self, _nodeCfg: Config):
        # the hidden list
        self._contract_dict = dict()
        self._sol_list = list()
        # publicly accessible
        self.project_workspace_root = ""
        self.accountAddr = None
        self.pathfinder = None
        self.artifact_manager = None
        self._sol_link = None
        self.is_deploy = False
        self.is_internal = False
        self.is_forge = False
        self.is_flashbot_prc = False
        self.deployed_address = False
        self.last_class = ""
        self.list_type = "list_address"
        self.network_cfg = _nodeCfg

        self.w3 = web3_provider(_nodeCfg.rpc_url)

        if "flastbot" in _nodeCfg.rpc_url:
            self.is_flashbot_prc = True

        self._optimizations = 200
        result = self.w3.isConnected()
        if not result:
            print(f"try to connect {self.network_cfg.network_name}  {Bolors.WARNING} {self.network_cfg.rpc_url}: {result} {Bolors.RESET}")
            exit(0)
            return
        else:
            print(f"You are now connected to {Bolors.OK} {self.network_cfg.network_name} {self.network_cfg.rpc_url} {Bolors.RESET}")

    def withPOA(self) -> "MiliDoS":
        """
        the normal usual term to fix some POA related problems
        :return:
        """
        self.w3.middleware_onion.inject(geth_poa_middleware, layer=0)
        return self

    def withFlastBot(self) -> "MiliDoS":

        return self

    def isAddress(self, address: str) -> bool:
        """
        Verification of the valid EVM address
        :param address:
        :return:
        """
        return self.w3.isAddress(address)

    def connect(self, workspace: str, history: any) -> None:

        """
        connect the existing deployed contract
        :param workspace: the workspace directory
        :param history: the deployed history folder under the deploy_history
        :return:
        """
        self.is_deploy = False
        self.artifact_manager = SolWeb3Tool()
        if history is False:
            self.pathfinder = Paths(workspace).setDefaultPath().Network(self.network_cfg.network_name)
        else:
            self.pathfinder = Paths(workspace).SetUseHistory(history).Network(self.network_cfg.network_name)

        self.ready_io(True)

    def SetupContract(self):
        pass

    def after_deployment_initialize_settings(self):
        """
        setup contract starting params
        setup the starting time using bang
        setup the first member
        :return:
        """
        pass

    def setWorkspace(self, path: str, readio: bool = True) -> "MiliDoS":
        self.project_workspace_root = path
        self.artifact_manager = SolWeb3Tool()
        self.pathfinder = Paths(path).setDefaultPath().Network(self.network_cfg.network_name)
        if readio:
            self.ready_io(True)
        return self

    def setClassSolNames(self, to_compile_contract_list: list) -> "MiliDoS":
        self._sol_list = to_compile_contract_list
        return self

    def setClassSolLinks(self, compile_links: list) -> "MiliDoS":
        self._sol_link = compile_links
        return self

    def setEvm(self, version_evm: str) -> "MiliDoS":
        """
        the specify the version of the ethereum virtual machine
        :param version_evm: the version of the EVM
        :return:
        """
        self.EVM_VERSION = version_evm
        return self

    def setOptimizationRuns(self, runs: int) -> "MiliDoS":
        self._optimizations = runs
        return self

    def remoteCompile(self, ver: str) -> "MiliDoS":
        """
        all parameters will be inserted automatically according to the previous setup
        :param ver:
        :return:
        """
        if ver == "":
            print("there is no solidity version specified")
            exit(0)
        self.pathfinder.setSolVersion(ver)
        self.pathfinder.setEvm(self.EVM_VERSION)
        BuildRemoteLinuxCommand(self.pathfinder, self._optimizations, self._sol_list, self._sol_link)
        return self

    def useForge(self) -> "MiliDoS":
        # ==================================================
        if self._sol_list is not None:
            for v in self._sol_list:
                based_name = os.path.basename(v)
                class_name = based_name.replace(".sol", "")
                # class_name_process = filter_file_name(based_name).replace('.sol', '')
                self.artifact_manager.SplitForgeBuild(class_name)
        self.is_forge = True
        return self

    def localTranspile(self, dapp_ts_folder: str = None) -> "MiliDoS":
        """
        :param dapp_ts_folder: the destination is follow by this path {dapp_ts_folder}/src/api/abi/xxx.ts
        if this valuable is None then there will not be any copy files to the destination
        :return: instance of moody
        """
        self.pathfinder.updateTargetDappFolder(dapp_ts_folder)
        if self.is_forge:
            BuildLangForge(self.pathfinder, self._sol_list)
        else:
            BuildLang(self.pathfinder, self._sol_list)
        return self

    def get_block(self, block_identifier, full_transactions: bool = False):
        """
        to see the block information
        :param block_identifier:
        :param full_transactions:
        :return: instance of moody
        """
        with w3_lock:
            res = self.w3.eth.getBlock(block_identifier, full_transactions)
        return res

    def erc20_contract(self):
        cTool = SolWeb3Tool()
        cTool.setBuildNameSpace("artifact").GetCodeClassFromBuild("ERC20")
        return self.w3.eth.contract(abi=cTool.abi)

    def estimate_gas_price(self):
        return self.w3.eth.gasPrice

    def send_contract_tx(self, contract: Web3Contract, function_name: str, from_acc: str,
                         private_key: bytes, gas: int = 0, gas_price: int = 0, _value: int = 0,
                         args: Tuple = ()):
        """
        Creates the contract tx and signs it with private_key to be transmitted as raw tx
        """

        tx = getattr(contract.functions, function_name)(*args).buildTransaction(
            {
                'from': from_acc,
                'chainId': self.w3.eth.chainId,
                # gas_price is in gwei
                'gasPrice': gas_price * 1e9 if gas_price else self.estimate_gas_price(),
                'gas': gas or None,
                'nonce': self.w3.eth.getTransactionCount(from_acc, block_identifier='pending'),
                'value': _value
            })
        signed_txn = self.w3.eth.account.sign_transaction(tx, private_key)
        return self.w3.eth.sendRawTransaction(signed_txn.rawTransaction)

    def contract_event_in_range(self, contract, event_name: str, from_block: int = 0, to_block: Optional[int] = None) -> None:
        """
        scans the blockchain, and yields blocks that has contract tx with the provided event
        Note: Be cautions with the range provided, as the logic creates query for each block which could be a bottleneck.
        :param from_block: starting block, defaults to 0
        :param to_block: end block, defaults to 'latest'
        :param provider:
        :param logger:
        :param contract:
        :param event_name: name of the contract emit event you wish to be notified of
        """
        if to_block is None:
            to_block = self.w3.eth.blockNumber

        with w3_lock:

            if isinstance(self.w3.provider, HTTPProvider):
                for block_num in range(from_block, to_block + 1):
                    block = self.w3.eth.getBlock(block_num, full_transactions=True)
                    contract_transactions = extract_tx_by_address(contract.address, block)

                    if not contract_transactions:
                        continue
                    for tx in contract_transactions:
                        _, log = event_log(tx_hash=tx.hash, events=[event_name], provider=self.w3, contract=contract.tracked_contract)
                        if log is None:
                            continue
                        yield log
            else:
                event = getattr(contract.tracked_contract.events, event_name)
                event_filter = event.createFilter(fromBlock=from_block, toBlock=to_block)

                for tx in event_filter.get_new_entries():
                    _, log = event_log(tx_hash=tx.hash, events=[event_name], provider=self.w3, contract=contract.tracked_contract)

                    if log is None:
                        continue

                    yield log

    def AuthByMemo(self, phrase: str = None) -> "MiliDoS":
        keyLo = self.w3.eth.account.from_mnemonic(phrase)
        # self.w3.eth.defaultAccount = keyoo.address
        self.w3.eth.account = keyLo
        # self.w3.eth.get_transaction_count
        # self.w3.eth.accounts[0] = keyLo.address
        # self.w3.eth.defaultAccount(f"0x{keyLo.key}")
        is_address = self.w3.isAddress(keyLo.address)
        # self.w3.isChecksumAddress(keyLo.address)
        self.accountAddr = keyLo.address
        print(f"🔫 You are now using {keyLo.address} and it is a {'valid key' if is_address else 'invalid key'}")

        return self

    def Auth(self, private_key_line: str = None) -> "MiliDoS":
        """
        switching the operating address to a different one that is given by the private key
        :param private_key_line: the input private key
        :return:
        """
        if private_key_line is None:
            private_key_line = DefaultKeys.k0

        # f"0x{private_key_line}"
        keyLo = self.w3.eth.account.from_key(f"0x{private_key_line}")
        # self.w3.eth.defaultAccount = keyoo.address
        self.w3.eth.account = keyLo
        # self.w3.eth.get_transaction_count
        # self.w3.eth.accounts[0] = keyLo.address
        # self.w3.eth.defaultAccount(f"0x{keyLo.key}")
        is_address = self.w3.isAddress(keyLo.address)
        # self.w3.isChecksumAddress(keyLo.address)
        self.accountAddr = keyLo.address
        print(f"🔫 You are now using {keyLo.address} and it is a {'valid key' if is_address else 'invalid key'}")
        if self.is_flashbot_prc:
            flashbot(self.w3, keyLo, self.network_cfg.rpc_url)

        return self

    def estimateGas(self, class_name: str) -> int:
        """
        only for testing the contract deploy gas requirement
        :param class_name:
        :return:
        """
        # estimate_gas
        solc_artifact = SolWeb3Tool()
        solc_artifact.setBasePath(self.project_workspace_root)
        if self.is_forge:
            solc_artifact = solc_artifact.GetCodeClassFromForgeBuild(class_name)
        else:
            solc_artifact = solc_artifact.GetCodeClassFromBuild(class_name)
        nr = self.w3.eth.contract(abi=solc_artifact.abi, bytecode=solc_artifact.bin)
        gas_est_amount = nr.constructor().estimateGas()
        price = self.w3.eth.generate_gas_price()
        # source: https://ethereum.stackexchange.com/questions/84943/what-is-the-equivalent-of-buildtransaction-of-web3py-in-web3js
        print(f"Price: {price}")
        return gas_est_amount

    def OverrideGasConfig(self, gas: int, gas_price: int) -> None:
        """
        the override the configuration for the gas amount and the gas price
        :param gas: int
        :param gas_price: int
        :return: NONE
        """
        self.network_cfg.gas = gas
        self.network_cfg.gasPrice = gas_price

    def OverrideChainConfig(self, one: int, wait: int) -> None:
        """
        Lets have the configuration done now.
        :param one: ONE coin to measure
        :param wait: the waiting time from each block confirmation
        :return:
        """
        self.network_cfg.wait_time = wait
        self.network_cfg.one = one

    @property
    def gas(self) -> int:
        return self.network_cfg.gas

    @property
    def gasPrice(self) -> int:
        return self.network_cfg.gasPrice

    @property
    def one(self) -> int:
        """
        ONE platform coin will be decoded to be...
        :return: int
        """
        return self.network_cfg.one

    @property
    def waitSec(self) -> int:
        return self.network_cfg.wait_time

    @property
    def LinkVRFHashKey(self) -> str:
        if self.network_cfg.link_keyhash is None:
            raise ValueError("Link VRF Hash Key is endorsed on this network")
        else:
            return self.network_cfg.link_keyhash

    @property
    def LinkVRFCoordinator(self) -> str:
        if self.network_cfg.link_vrf_coordinator is None:
            raise ValueError("Link VRF is endorsed on this network")
        else:
            return self.network_cfg.link_vrf_coordinator

    @property
    def LinkTokenAddress(self) -> str:
        if self.network_cfg.link_token is None:
            raise ValueError("Link Token is endorsed on this network")
        else:
            return self.network_cfg.link_token

    def _checkErrorForTxReceipt(self, receipt: any, class_name: str, jsonfile: str) -> None:
        if "contractAddress" not in receipt:
            print(f"⚠️ Error from deploy contract and no valid address found for {class_name}.")
            raise InvalidAddress

        if "transactionHash" not in receipt:
            print(f"⚠️ The deployment is failed because there is no valid address found from {class_name}. Please check for internal errors from deployment has {receipt.transactionHash}")
            raise InvalidAddress

        hash = str(receipt.transactionHash)
        preaddress = str(receipt.contractAddress)
        if self.isAddress(preaddress) is False:
            print(f"⚠️ The deployment is failed because there is no valid address found from {class_name}. Please check for internal errors from deployment hash from {jsonfile}")
            raise InvalidAddress

    def provide_artifact_extends(self, class_name: str) -> SolWeb3Tool:
        """
        Following the class name of the contract
        :param class_name:
        :return:
        """
        if not self.artifact_manager:
            print("❌ Root path is not setup. please setup the workspace first.")
            exit(2)

        sol = self.artifact_manager
        sol.setBasePath(self.project_workspace_root)
        sol.setBuildNameSpace("build")

        if self.is_forge:
            sol = sol.GetCodeClassFromForgeBuild(class_name)
        else:
            sol = sol.GetCodeClassFromBuild(class_name)

        self.artifact_manager = sol
        return sol

    def provide_artifact_implemented(self, class_name: str) -> SolWeb3Tool:
        """
        Use the internal available class names. Please see the internal class name list
        :param class_name:
        :return:
        """
        if not self.artifact_manager:
            print("❌ Root path is not setup. please setup the workspace first.")
            exit(2)

        sol = self.artifact_manager
        sol.setBasePath(root_base_path)
        sol.setBuildNameSpace("artifacts")
        sol = sol.GetCodeClassFromBuild(class_name)
        self.artifact_manager = sol
        return sol

    def deployImple(self, class_name: str, params: list = [], gas_price: int = 0, gas_limit: int = 0) -> bool:
        """
        Deployment of implemented abi and code
        :param class_name:
        :param params:
        :param gas_price:
        :param gas_limit:
        :return:
        """
        contract_nv = None
        try:
            solc_artifact = self.provide_artifact_implemented(class_name)
            bin = BinOp(solc_artifact.bin, class_name)
            if bin.checkBinForUndeployLib() is True:
                bin.setDebug(True)
                # try to find the needed libraries in address..
                bin.anaylze(self)
                contract_nv = self.w3.eth.contract(abi=solc_artifact.abi, bytecode=bin.GetKnifedBin())
            else:
                contract_nv = self.w3.eth.contract(abi=solc_artifact.abi, bytecode=bin.GetRawBin())

        except FileNotFoundError:
            print("💢 bin or abi file is not found.")
            exit(3)
        except FoundUndeployedLibraries:
            exit(4)
        except ContractLogicError as e:
            print(f"💢 Contract error {e}")
            exit(5)
        gasprice = self.gasPrice if gas_price == 0 else gas_price
        gas = self.gas if gas_limit == 0 else gas_limit
        if len(params) > 0:
            _transaction = contract_nv.constructor(*params).buildTransaction({
                "gasPrice": gasprice,
                "gas": gas
            })
        else:
            _transaction = contract_nv.constructor().buildTransaction({
                "gasPrice": gasprice,
                "gas": gas
            })
        self.artifact_manager.setBasePath(self.project_workspace_root)
        return self._endingdeployment(_transaction, class_name)

    def deploy(self, class_name: str, params: list = [], gas_price: int = 0, gas_limit: int = 0) -> bool:
        """
        This is using the faster way to deploy files by using the specific abi and bin files.
        If all these parameters to be ignored then these things will be taken from other available
        valuables for gas price and gas limit
        :param class_name: the input class name
        :param params: the parameters
        :param gas_price: the gas price
        :param gas_limit: the gas limit
        :return:
        """
        contract_nv = None
        try:
            solc_artifact = self.provide_artifact_extends(class_name)
            bin = BinOp(solc_artifact.bin, class_name)
            if bin.checkBinForUndeployLib() is True:
                bin.setDebug(True)
                # try to find the needed libraries in address..
                bin.anaylze(self)
                contract_nv = self.w3.eth.contract(abi=solc_artifact.abi, bytecode=bin.GetKnifedBin())
            else:
                contract_nv = self.w3.eth.contract(abi=solc_artifact.abi, bytecode=bin.GetRawBin())

        except FileNotFoundError:
            print("💢 bin or abi file is not found.")
            exit(3)
        except FoundUndeployedLibraries:
            exit(4)
        except ContractLogicError as e:
            print(f"💢 Contract error {e}")
            exit(5)

        gasprice = self.gasPrice if gas_price == 0 else gas_price
        gas = self.gas if gas_limit == 0 else gas_limit
        if len(params) > 0:
            _transaction = contract_nv.constructor(*params).buildTransaction({
                "gasPrice": gasprice,
                "gas": gas
            })
        else:
            _transaction = contract_nv.constructor().buildTransaction({
                "gasPrice": gasprice,
                "gas": gas
            })
        return self._endingdeployment(_transaction, class_name)

    def _endingdeployment(self, _transaction: any, class_name: str) -> bool:
        try:
            _transaction['nonce'] = self.w3.eth.getTransactionCount(self.accountAddr)
            _transaction['to'] = None
            # _transaction['gas'] = self.gas if gas_limit == 0 else gas_limit
            # _transaction['gasPrice'] = self.gasPrice if gas_price == 0 else gas_price
            # _transaction['gas'] = 2200000000,
            print("ok --- ", _transaction)
            # Get correct transaction nonce for sender from the node
            print(f"========🖍 Signing {class_name}, gas:{_transaction['gas']}, price:{_transaction['gasPrice']} ...")
            signed = self.w3.eth.account.sign_transaction(_transaction)

            txHash = self.w3.eth.sendRawTransaction(signed.rawTransaction)
            # print(f"Contract '{class_name}' deployed; Waiting to transaction receipt")
            print(f"========Wait for Block Confirmation - {class_name} ☕️")
            tx_receipt = self.w3.eth.waitForTransactionReceipt(txHash)
            print("========TX Pre-Result ✅")
            print(tx_receipt)
            print(f"========Broadcast Result ✅ -> {Paths.showCurrentDeployedClass(class_name)}")

            self._checkErrorForTxReceipt(tx_receipt, class_name, Paths.showCurrentDeployedClass(class_name))
            fresh_address = tx_receipt.contractAddress
            self._contract_dict[class_name] = fresh_address
            self.deployed_address = fresh_address
            self.setTargetClass(class_name)
            self.setKV("by", self.accountAddr)
            print("📦 Address saved to ✅ {} -> {}".format(fresh_address, class_name))
            print(f"🔍 You can check with the explorer for more detail: {Bolors.WARNING} {self.network_cfg.block_explorer}{Bolors.RESET}")

            self.artifact_manager.StoreTxResult(tx_receipt, self.pathfinder.classObject(class_name))
            self.complete_deployment()
            return True
        except InvalidAddress:
            return False
        except ContractLogicError as e:
            print(f"Error: {e}")
            return False
        except TimeExhausted:
            print("After 120 seconds, the boardcast block is not in the chain.")
            return False
        except ValueError as te:
            if "code" in te:
                code = int(te["code"])
                if code == -32000:
                    print("NOT ENOUGH GAS - insufficient funds for gas")
                    return False
            print(te)
            return False

    @property
    def __list_key_label(self) -> str:
        return "{}_{}".format(self.list_type, self.last_class)

    @property
    def __kv_label(self) -> str:
        return "kv_{}".format(self.last_class)

    def getAddr(self, keyname: str) -> str:
        """example: TT67rPNwgmpeimvHUMVzFfKsjL9GZ1wGw8"""
        return self._contract_dict.get(keyname)

    def getAllAddress(self) -> dict:
        return self._contract_dict

    def preview_all_addresses(self) -> None:
        print(self._contract_dict)

    def is_deployment(self) -> bool:
        return self.is_deploy

    def ready_io(self, show_address: bool = False):
        """try to load up the file from the existing path"""
        try:
            self._contract_dict = self.pathfinder.LoadDeploymentFile()
            print("📦 Review the loaded deployment data from ... ")
            if show_address:
                self.preview_all_addresses()
        except FileNotFoundError:
            print("💢 Deployment File is not found ...")
        except TypeError as e:
            print(e)

    def setTargetClass(self, classname: str) -> "MiliDoS":
        self.last_class = classname
        return self

    def setTargetListName(self, listname: str) -> "MiliDoS":
        self.list_type = listname
        return self

    def setKV(self, key: str, value: any) -> "MiliDoS":
        if self.__kv_label not in self._contract_dict:
            self._contract_dict[self.__kv_label] = dict()
        self._contract_dict[self.__kv_label][key] = value
        return self

    def hasAddressInList(self, address: str) -> bool:
        if self.__list_key_label not in self._contract_dict:
            return False
        try:
            v = self._contract_dict[self.__list_key_label].index(address)
            return True
        except ValueError:
            return False

    def pushAddress(self, address: str, unique: bool = True) -> bool:
        if self.__list_key_label not in self._contract_dict:
            self._contract_dict[self.__list_key_label] = list()

        if unique is True:
            try:
                found_index = self._contract_dict[self.__list_key_label].index(address)
                return False
            except ValueError:
                self._contract_dict[self.__list_key_label].append(address)
                return True
            except IndexError:
                self._contract_dict[self.__list_key_label].append(address)
                return True
        else:
            self._contract_dict[self.__list_key_label].append(address)
            return True

    def removeAddress(self, address: str) -> bool:
        if self.__list_key_label not in self._contract_dict:
            return False
        self._contract_dict[self.__list_key_label].remove(address)
        return True

    def iterList(self) -> iter:
        if self.__list_key_label not in self._contract_dict:
            raise Exception("there is no list in the map")
        return iter(self._contract_dict[self.__list_key_label])

    def hasList(self) -> bool:
        if self.__list_key_label not in self._contract_dict:
            return False
        return len(self._contract_dict[self.__list_key_label]) > 0

    def hasField(self, key: str) -> bool:
        if self.__kv_label not in self._contract_dict:
            self._contract_dict[self.__kv_label] = dict()

        if key not in self._contract_dict[self.__kv_label]:
            return False
        else:
            return True

    def hasContractName(self, name: str) -> bool:
        return name in self._contract_dict

    def getString(self, key: str) -> str:
        return str(self.getVal(key))

    def getInt(self, key: str) -> int:
        return int(self.getVal(key))

    def getBytesArray(self, key: str) -> bytearray:
        return bytearray(self.getVal(key))

    def getBytes(self, key: str) -> bytes:
        return bytes(self.getVal(key))

    def getFloat(self, key: str) -> float:
        return float(self.getVal(key))

    def getVal(self, key: str) -> any:
        if self.__kv_label not in self._contract_dict:
            self._contract_dict[self.__kv_label] = dict()

        if key in self._contract_dict[self.__kv_label]:
            return self._contract_dict[self.__kv_label][key]

        return ""

    def complete_deployment(self) -> None:
        """store up the deployed contrcat addresses to the local file storage"""
        self.artifact_manager.StoreTxResult(self._contract_dict, self.pathfinder.SaveDeployConfig)

    def SaveConfig(self) -> None:
        self.complete_deployment()



"""Generated wrapper for Eth2MultiDeposit Solidity contract."""

# pylint: disable=too-many-arguments

import json
from typing import (  # pylint: disable=unused-import
    Any,
    List,
    Optional,
    Tuple,
    Union,
)
import time
from eth_utils import to_checksum_address
from mypy_extensions import TypedDict  # pylint: disable=unused-import
from hexbytes import HexBytes
from web3 import Web3
from web3.contract import ContractFunction
from web3.datastructures import AttributeDict
from web3.providers.base import BaseProvider
from web3.exceptions import ContractLogicError
from moody.m.bases import ContractMethod, Validator, ContractBase, Signatures
from moody.m.tx_params import TxParams
from moody.libeb import MiliDoS
from moody import Bolors

# Try to import a custom validator class definition; if there isn't one,
# declare one that we can instantiate for the default argument to the
# constructor for Eth2MultiDeposit below.
try:
    # both mypy and pylint complain about what we're doing here, but this
    # works just fine, so their messages have been disabled here.
    from . import (  # type: ignore # pylint: disable=import-self
        Eth2MultiDepositValidator,
    )
except ImportError:

    class Eth2MultiDepositValidator(  # type: ignore
        Validator
    ):
        """No-op input validator."""

try:
    from .middleware import MIDDLEWARE  # type: ignore
except ImportError:
    pass


class DepositMethod(ContractMethod):  # pylint: disable=invalid-name
    """Various interfaces to the deposit method."""

    def __init__(self, elib: MiliDoS, contract_address: str, contract_function: ContractFunction, validator: Validator = None):
        """Persist instance data."""
        super().__init__(elib, contract_address, validator)
        self._underlying_method = contract_function
        self.sign = validator.getSignature("deposit")

    def validate_and_normalize_inputs(self, pubkey: List[Union[bytes, str]], withdrawal_credentials: List[Union[bytes, str]], signature: List[Union[bytes, str]], deposit_data_root: List[Union[bytes, str]]) -> any:
        """Validate the inputs to the deposit method."""
        self.validator.assert_valid(
            method_name='deposit',
            parameter_name='pubkey',
            argument_value=pubkey,
        )
        self.validator.assert_valid(
            method_name='deposit',
            parameter_name='withdrawal_credentials',
            argument_value=withdrawal_credentials,
        )
        self.validator.assert_valid(
            method_name='deposit',
            parameter_name='signature',
            argument_value=signature,
        )
        self.validator.assert_valid(
            method_name='deposit',
            parameter_name='deposit_data_root',
            argument_value=deposit_data_root,
        )
        return (pubkey, withdrawal_credentials, signature, deposit_data_root)

    def block_send(self, pubkey: List[Union[bytes, str]], withdrawal_credentials: List[Union[bytes, str]], signature: List[Union[bytes, str]], deposit_data_root: List[Union[bytes, str]], _valeth: int = 0) -> None:
        """Execute underlying contract method via eth_call.

        :param tx_params: transaction parameters
        :returns: the return value of the underlying method.
        """
        _fn = self._underlying_method(pubkey, withdrawal_credentials, signature, deposit_data_root)
        try:

            _t = _fn.buildTransaction({
                'from': self._operate,
                'gas': self.gas_limit,
                'gasPrice': self.gas_price_wei
            })
            _t['nonce'] = self._web3_eth.getTransactionCount(self._operate)

            if _valeth > 0:
                _t['value'] = _valeth

            if self.debug_method:
                print(f"======== Signing ✅ by {self._operate}")
                print(f"======== Transaction ✅ check")
                print(_t)

            if 'data' in _t:

                signed = self._web3_eth.account.sign_transaction(_t)
                txHash = self._web3_eth.sendRawTransaction(signed.rawTransaction)
                tx_receipt = None
                if self.auto_reciept is True:
                    print(f"======== awaiting Confirmation 🚸️ {self.sign}")
                    tx_receipt = self._web3_eth.wait_for_transaction_receipt(txHash)
                    if self.debug_method:
                        print("======== TX Result ✅")
                        print(tx_receipt)

                self._on_receipt_handle("deposit", tx_receipt, txHash)

            if self.auto_reciept is False:
                time.sleep(self._wait)


        except ContractLogicError as er:
            print(f"{Bolors.FAIL}Error {er} {Bolors.RESET}: deposit")
            message = f"Error {er}: deposit"
            self._on_fail("deposit", message)
        except ValueError as err:
            if "message" in err.args[0]:
                message = err.args[0]["message"]
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, deposit: {message}")
            else:
                message = "Error Revert , Reason: Unknown"
                print(f"{Bolors.FAIL}Error Revert {Bolors.RESET}, deposit. Reason: Unknown")

            self._on_fail("deposit", message)

    def send_transaction(self, pubkey: List[Union[bytes, str]], withdrawal_credentials: List[Union[bytes, str]], signature: List[Union[bytes, str]], deposit_data_root: List[Union[bytes, str]], tx_params: Optional[TxParams] = None) -> Union[HexBytes, bytes]:
        """Execute underlying contract method via eth_sendTransaction.

        :param tx_params: transaction parameters
        """
        (pubkey, withdrawal_credentials, signature, deposit_data_root) = self.validate_and_normalize_inputs(pubkey, withdrawal_credentials, signature, deposit_data_root)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(pubkey, withdrawal_credentials, signature, deposit_data_root).transact(tx_params.as_dict())

    def build_transaction(self, pubkey: List[Union[bytes, str]], withdrawal_credentials: List[Union[bytes, str]], signature: List[Union[bytes, str]], deposit_data_root: List[Union[bytes, str]], tx_params: Optional[TxParams] = None) -> dict:
        """Construct calldata to be used as input to the method."""
        (pubkey, withdrawal_credentials, signature, deposit_data_root) = self.validate_and_normalize_inputs(pubkey, withdrawal_credentials, signature, deposit_data_root)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(pubkey, withdrawal_credentials, signature, deposit_data_root).buildTransaction(tx_params.as_dict())

    def estimate_gas(self, pubkey: List[Union[bytes, str]], withdrawal_credentials: List[Union[bytes, str]], signature: List[Union[bytes, str]], deposit_data_root: List[Union[bytes, str]], tx_params: Optional[TxParams] = None) -> int:
        """Estimate gas consumption of method call."""
        (pubkey, withdrawal_credentials, signature, deposit_data_root) = self.validate_and_normalize_inputs(pubkey, withdrawal_credentials, signature, deposit_data_root)
        tx_params = super().normalize_tx_params(tx_params)
        return self._underlying_method(pubkey, withdrawal_credentials, signature, deposit_data_root).estimateGas(tx_params.as_dict())


class SignatureGenerator(Signatures):
    """
        The signature is generated for this and it is installed.
    """

    def __init__(self, abi: any):
        super().__init__(abi)

    def deposit(self) -> str:
        return self._function_signatures["deposit"]


# pylint: disable=too-many-public-methods,too-many-instance-attributes
class Eth2MultiDeposit(ContractBase):
    """Wrapper class for Eth2MultiDeposit Solidity contract."""
    _fn_deposit: DepositMethod
    """Constructor-initialized instance of
    :class:`DepositMethod`.
    """

    SIGNATURES: SignatureGenerator = None

    def __init__(
            self,
            core_lib: MiliDoS,
            contract_address: str,
            validator: Eth2MultiDepositValidator = None,
    ):
        """Get an instance of wrapper for smart contract.
        """
        # pylint: disable=too-many-statements
        super().__init__(contract_address, Eth2MultiDeposit.abi())
        web3 = core_lib.w3

        if not validator:
            validator = Eth2MultiDepositValidator(web3, contract_address)

        # if any middleware was imported, inject it
        try:
            MIDDLEWARE
        except NameError:
            pass
        else:
            try:
                for middleware in MIDDLEWARE:
                    web3.middleware_onion.inject(
                        middleware['function'], layer=middleware['layer'],
                    )
            except ValueError as value_error:
                if value_error.args == ("You can't add the same un-named instance twice",):
                    pass

        self._web3_eth = web3.eth
        functions = self._web3_eth.contract(address=to_checksum_address(contract_address), abi=Eth2MultiDeposit.abi()).functions
        self._signatures = SignatureGenerator(Eth2MultiDeposit.abi())
        validator.bindSignatures(self._signatures)

        self._fn_deposit = DepositMethod(core_lib, contract_address, functions.deposit, validator)

    def deposit(self, pubkey: List[Union[bytes, str]], withdrawal_credentials: List[Union[bytes, str]], signature: List[Union[bytes, str]], deposit_data_root: List[Union[bytes, str]], wei: int = 0) -> None:
        """
        Implementation of deposit in contract Eth2MultiDeposit
        Method of the function
    
        """

        self._fn_deposit.callback_onfail = self._callback_onfail
        self._fn_deposit.callback_onsuccess = self._callback_onsuccess
        self._fn_deposit.auto_reciept = self.call_contract_enforce_tx_receipt
        self._fn_deposit.gas_limit = self.call_contract_fee_amount
        self._fn_deposit.gas_price_wei = self.call_contract_fee_price
        self._fn_deposit.debug_method = self.call_contract_debug_flag

        self._fn_deposit.wei_value = wei

        return self._fn_deposit.block_send(pubkey, withdrawal_credentials, signature, deposit_data_root, wei)

    def CallContractWait(self, t_long: int) -> "Eth2MultiDeposit":
        self._fn_deposit.setWait(t_long)
        return self

    @staticmethod
    def abi():
        """Return the ABI to the underlying contract."""
        return json.loads(
            '[{"inputs":[{"internalType":"bytes[]","name":"pubkey","type":"bytes[]"},{"internalType":"bytes[]","name":"withdrawal_credentials","type":"bytes[]"},{"internalType":"bytes[]","name":"signature","type":"bytes[]"},{"internalType":"bytes32[]","name":"deposit_data_root","type":"bytes32[]"}],"name":"deposit","outputs":[],"stateMutability":"payable","type":"function"}]'  # noqa: E501 (line-too-long)
        )

# pylint: disable=too-many-lines



#!/usr/bin/env python
"""Base wrapper class for accessing ethereum smart contracts."""
import codecs
import json
import os

# from datetime import datetime
# from typing import Any, Union, Tuple
from moody import Evm


class Paths:
    """
    :type statement
    Manage the workspace paths
    It takes cared the paths related primitives

    """
    statement = 'End : {}, IO File {}'
    _contract_dict: dict
    _network: str
    FILE_CONTRACT = "backedup"
    ACTION_FOLDER = "deploy_results"
    HISTORY_FOLDER = "deploy_history"
    COLLECTION_CONTRACTS = "players"
    DEPLOYMENT_FILE_NAME = "deploy_{}{}.json"
    VERSION_NAME = "v{}"
    NAME_FILE_EXX = "{}/{}.json"
    # ---- sovernior
    TARGET_LOC = "/root/contracts"
    # ---- including the local bash file
    LOCAL_BASH_INCLUDE = "inc"
    COMPRESSED_NAME = "solc-build.tar.gz"
    SOLC_VER = "0.5.15"
    SOLCPATH = "/root/contracts/vault"
    EVM_VERSION = Evm.ISTANBUL
    BUILDPATH = ""
    WEB_DAPP_SRC = None
    SUB_FIX = ""

    def __init__(self, root_path_as_workspace):
        self.___workspace = root_path_as_workspace
        self.___current_deployment_path = os.path.join(self.___workspace, self.ACTION_FOLDER)
        self.___network_name = "mainnet"
        self.BUILDPATH = root_path_as_workspace

    @property
    def subFix(self) -> str:
        """preview the file name"""
        return self.SUB_FIX

    @subFix.setter
    def subFix(self, sub: str) -> "Paths":
        """the file name does not require extension name"""
        self.SUB_FIX = sub
        return self

    @classmethod
    def showCurrentDeployedClass(cls, class_name: str) -> str:
        return cls.NAME_FILE_EXX.format(cls.ACTION_FOLDER, class_name)

    def updateTargetDappFolder(self, folderName: str) -> "Paths":
        self.WEB_DAPP_SRC = folderName
        return self

    def setSolVersion(self, version: str) -> "Paths":
        self.SOLC_VER = version
        return self

    def setEvm(self, ver: str) -> "Paths":
        self.EVM_VERSION = ver
        return self

    def setBashInclude(self, p: str) -> "Paths":
        self.LOCAL_BASH_INCLUDE = p
        return self

    def setDefaultPath(self) -> "Paths":
        self.___current_deployment_path = os.path.join(self.___workspace, self.ACTION_FOLDER)
        return self

    def workspaceFilename(self, name) -> str:
        return os.path.join(self.___workspace, name)

    def classObject(self, className: str) -> str:
        return os.path.join(self.___current_deployment_path, "{}.json".format(className))

    @property
    def __playerAddrsFilePath(self) -> str:
        return os.path.join(self.___current_deployment_path, "{}.json".format(self.COLLECTION_CONTRACTS))

    @property
    def __deploymentPath(self) -> str:
        return os.path.join(self.___current_deployment_path, self.DEPLOYMENT_FILE_NAME.format(self.___network_name, self.subFix))

    @property
    def SaveDeployConfig(self) -> str:
        return self.__deploymentPath

    @property
    def SavePlayersList(self) -> str:
        return self.__playerAddrsFilePath

    """
    :type Config the network name
    """

    def Network(self, name) -> "Paths":
        self.___network_name = name
        return self

    def SetUseHistory(self, history_path: str) -> "Paths":
        self.___current_deployment_path = os.path.join(self.___workspace, self.HISTORY_FOLDER, history_path)
        return self

    def SetUseVersion(self, version_name: str) -> "Paths":
        version = self.VERSION_NAME.format(version_name)
        self.SetUseHistory(version)
        return self

    def LoadDeploymentFile(self) -> dict:
        return json.load(codecs.open(self.__deploymentPath, 'r', 'utf-8-sig'))

    def LoadPlayerFile(self) -> dict:
        return json.load(codecs.open(self.__playerAddrsFilePath, 'r', 'utf-8-sig'))



import os
import re

from ..paths import Paths
from . import ITEM_CP_LOCAL, ITEM_TRANSPILE_GO, TRANS_LOCAL, ITEM_TRANSPILE_PYTHON, ITEM_TRANSPILE_TS, PRE_HEAD, SUB_FOOTER

REG = r"(.+?)([A-Z])"


def snake(match):
    return match.group(1).lower() + "_" + match.group(2).lower()


def filter_file_name(y: str) -> str:
    classNameNew = y
    if y.startswith("TRC"):
        classNameNew = y.lower()
    elif y.startswith("ERC20"):
        classNameNew = y.upper()
    else:
        classNameNew = re.sub(REG, snake, y, 0)
    print(classNameNew)
    return classNameNew


def moveTsFiles(p: Paths, pathName: str) -> str:
    nameClass = filter_file_name(os.path.basename(pathName)).replace('.sol', '')
    fromp = "{}/codec/gen_ts/{}.ts".format(p.BUILDPATH, nameClass)
    top = "{}/{}/src/api/abi/{}.ts".format(p.BUILDPATH, p.WEB_DAPP_SRC, nameClass)
    return ITEM_CP_LOCAL.format(
        fromlocation=fromp,
        tolocation=top
    )


def abiPath_v1(p: Paths, pathName: str) -> str:
    return f"{p.BUILDPATH}/build/{os.path.basename(pathName).replace('.sol', '')}.abi"


def abiPath_v2(p: Paths, pathName: str) -> str:
    file_name = os.path.basename(pathName)
    return f"{p.BUILDPATH}/build/{file_name}/{file_name.replace('.sol', '')}.abi"


def buildCmdPy(p: Paths, pathName: str) -> str:
    return ITEM_TRANSPILE_PYTHON.format(
        outputfolder=f"{p.BUILDPATH}/codec/gen_py",
        target_abi=abiPath_v1(p, pathName),
        BUILDPATH=p.BUILDPATH
    )


def buildCmdPy2(p: Paths, pathName: str) -> str:
    return ITEM_TRANSPILE_PYTHON.format(
        outputfolder=f"{p.BUILDPATH}/codec/gen_py",
        target_abi=abiPath_v2(p, pathName),
        BUILDPATH=p.BUILDPATH
    )


def buildCmdTs(p: Paths, pathName: str) -> str:
    return ITEM_TRANSPILE_TS.format(
        outputfolder=f"{p.BUILDPATH}/codec/gen_ts",
        target_abi=abiPath_v1(p, pathName),
        BUILDPATH=p.BUILDPATH
    )


def buildCmdTs2(p: Paths, pathName: str) -> str:
    return ITEM_TRANSPILE_TS.format(
        outputfolder=f"{p.BUILDPATH}/codec/gen_ts",
        target_abi=abiPath_v2(p, pathName),
        BUILDPATH=p.BUILDPATH
    )


def buildCmdGo(p: Paths, pathName: str) -> str:
    based_name = os.path.basename(pathName)
    class_name = filter_file_name(based_name).replace('.sol', '')
    return ITEM_TRANSPILE_GO.format(
        outputfolder=f"{p.BUILDPATH}/codec/gen_go",
        target_abi=abiPath_v1(p, pathName),
        BUILDPATH=p.BUILDPATH,
        classname=class_name
    )


def buildCmdGo2(p: Paths, pathName: str) -> str:
    based_name = os.path.basename(pathName)
    class_name = filter_file_name(based_name).replace('.sol', '')
    return ITEM_TRANSPILE_GO.format(
        outputfolder=f"{p.BUILDPATH}/codec/gen_go",
        target_abi=abiPath_v2(p, pathName),
        BUILDPATH=p.BUILDPATH,
        classname=class_name
    )


def wrapContentTranspile(tar: Paths, compile_list: list) -> str:
    """
    wrap content
    :param tar: path in string
    :param compile_list: the list in compile
    :return:
    """
    head_section = PRE_HEAD.format(path_definitions=tar.LOCAL_BASH_INCLUDE)
    contract_list_content = "\n".join(compile_list)
    return TRANS_LOCAL.format(
        TARGET_LOC=tar.TARGET_LOC,
        COMPRESSED_NAME=tar.COMPRESSED_NAME,
        SOLVER=tar.SOLC_VER,
        LISTP=contract_list_content,
        PRE_HEAD=head_section,
        FOOTER=SUB_FOOTER
    )


def BuildLang(p: Paths, list_class_names: list) -> None:
    """

    :param p: path in string
    :param list_class_names: the class name
    :return:
    """
    k = list()
    # ==================================================
    for v in list_class_names:
        k.append(buildCmdPy(p, v))
        k.append(buildCmdTs(p, v))
        k.append(buildCmdGo(p, v))
        if p.WEB_DAPP_SRC is not None:
            k.append(moveTsFiles(p, v))
    # ==================================================
    with open(p.workspaceFilename("localpile"), 'w') as f:
        f.write(wrapContentTranspile(p, k))
        f.close()


def BuildLangForge(p: Paths, list_class_names: list) -> None:
    """

    :param p: path in string
    :param list_class_names: the class name
    :return:
    """
    k = list()
    # ==================================================
    for v in list_class_names:
        k.append(buildCmdPy2(p, v))
        k.append(buildCmdTs2(p, v))
        k.append(buildCmdGo2(p, v))
        if p.WEB_DAPP_SRC is not None:
            k.append(moveTsFiles(p, v))
    # ==================================================
    with open(p.workspaceFilename("localpile"), 'w') as f:
        f.write(wrapContentTranspile(p, k))
        f.close()



"""Base wrapper class for accessing ethereum smart contracts."""

from typing import Any

from eth_utils import is_address, to_checksum_address

from web3 import Web3

from .tx_params import TxParams
# from web3.providers.base import BaseProvider
from ..libeb import MiliDoS
from .. import Bolors


class Signatures:
    _function_signatures = {}

    def __init__(self, abi: any):
        for func in [obj for obj in abi if obj['type'] == 'function']:
            name = func['name']
            types = [input['type'] for input in func['inputs']]
            self._function_signatures[name] = '{}({})'.format(name, ','.join(types))
        self._abi_store = abi

    def fromSignatures(self) -> dict:
        return self._function_signatures


class Validator:
    """Base class for validating inputs to methods."""

    def __init__(
            self,
            web3_or_provider: Web3,
            contract_address: str,
    ):
        """Initialize the instance."""
        pass

    def assert_valid(
            self, method_name: str, parameter_name: str, argument_value: Any
    ):
        """Raise an exception if method input is not valid.

        :param method_name: Name of the method whose input is to be validated.
        :param parameter_name: Name of the parameter whose input is to be
            validated.
        :param argument_value: Value of argument to parameter to be validated.
        """

    def bindSignatures(self, so: Signatures) -> None:
        self._bind_singatures = so

    def getSignature(self, sign_name: str) -> str:
        return self._bind_singatures.fromSignatures()[sign_name]


class ContractMethod:
    """Base class for wrapping an Ethereum smart contract method."""

    def __init__(
            self,
            elib: MiliDoS,
            contract_address: str,
            validator: Validator = None,
    ):
        """Instantiate the object.

        :param provider: Instance of :class:`web3.providers.base.BaseProvider`
        :param contract_address: Where the contract has been deployed to.
        :param validator: Used to validate method inputs.
        """

        self._web3_eth = elib.w3.eth  # pylint: disable=no-member
        if validator is None:
            validator = Validator(self._web3_eth, contract_address)
        self.validator = validator
        self._operate = elib.accountAddr
        self._wait = 5
        self.wei_value = 0
        self.gas_limit = 0
        self.gas_price_wei = 0
        self.auto_reciept = False
        self.debug_method = False
        self.callback_onsuccess = None
        self.callback_onfail = None

    @staticmethod
    def validate_and_checksum_address(address: str):
        """Validate the given address, and return it's checksum address."""
        if not is_address(address):
            raise TypeError("Invalid address provided: {}".format(address))
        return to_checksum_address(address)

    def normalize_tx_params(self, tx_params) -> TxParams:
        """Normalize and return the given transaction parameters."""
        if not tx_params:
            tx_params = TxParams()
        if not tx_params.from_:
            tx_params.from_ = self._web3_eth.coinbase or (
                self._web3_eth.accounts[0]
                if len(self._web3_eth.accounts) > 0
                else None
            )
        if tx_params.from_:
            tx_params.from_ = self.validate_and_checksum_address(
                tx_params.from_
            )
        return tx_params

    def setWait(self, t: int) -> "ContractMethod":
        self._wait = t
        return self

    def _on_receipt_handle(self, method_name: str, receipt=None, boardcast_hash=None) -> None:
        print(f"======== TX blockHash ✅")
        if receipt is not None:
            print(f"{Bolors.OK}{receipt.blockHash.hex()}{Bolors.RESET}")
            if self.callback_onsuccess is not None:
                self.callback_onsuccess(receipt.blockHash.hex(), method_name)
        else:
            if boardcast_hash is not None:
                print(f"{Bolors.WARNING}{boardcast_hash.hex()}{Bolors.RESET} - broadcast hash")

    def _on_fail(self, name: str, message: str) -> None:
        if self.callback_onfail is not None:
            self.callback_onfail(name, message)


class ContractBase:

    def __init__(self, address: str, abi: any):
        self.call_contract_fee_amount: int = 2000000000
        self.call_contract_fee_price: int = 105910000000
        self.call_contract_debug_flag: bool = False
        self.call_contract_enforce_tx_receipt: bool = True
        self.contract_address = address
        self._callback_onsuccess = None
        self._callback_onfail = None
        self._abi_store = abi
        self._signatures = None

    @property
    def fromAbi(self) -> any:
        return self._abi_store

    def CallAutoConf(self, f: MiliDoS) -> "ContractBase":
        self.call_contract_fee_amount = f.gas
        self.call_contract_fee_price = f.gasPrice
        return self

    def CallContractFee(self, gas: int, price: int) -> "ContractBase":
        self.call_contract_fee_amount = gas
        self.call_contract_fee_price = price
        return self

    def CallDebug(self, yesno: bool) -> "ContractBase":
        self.call_contract_debug_flag = yesno
        return self

    def EnforceTxReceipt(self, yesno: bool) -> "ContractBase":
        self.call_contract_enforce_tx_receipt = yesno
        return self

    def onSuccssCallback(self, cb: any) -> "ContractBase":
        self._callback_onsuccess = cb
        return self

    def onFailCallback(self, cb: any) -> "ContractBase":
        self._callback_onfail = cb
        return self

    def CallSignature(self) -> Signatures:
        return self._signatures



from eth_account.signers.local import LocalAccount
from eth_typing import HexStr
from hexbytes import HexBytes
from typing import TypedDict, List, Union
from web3.types import TxParams, _Hash32

# unsigned transaction
FlashbotsBundleTx = TypedDict(
    "FlashbotsBundleTx",
    {
        "transaction": TxParams,
        "signer": LocalAccount,
    },
)

# signed transaction
FlashbotsBundleRawTx = TypedDict(
    "FlashbotsBundleRawTx",
    {
        "signed_transaction": HexBytes,
    },
)

# transaction dict taken from w3.eth.get_block('pending', full_transactions=True)
FlashbotsBundleDictTx = TypedDict(
    "FlashbotsBundleDictTx",
    {
        "accessList": list,
        "blockHash": HexBytes,
        "blockNumber": int,
        "chainId": str,
        "from": str,
        "gas": int,
        "gasPrice": int,
        "maxFeePerGas": int,
        "maxPriorityFeePerGas": int,
        "hash": HexBytes,
        "input": str,
        "nonce": int,
        "r": HexBytes,
        "s": HexBytes,
        "to": str,
        "transactionIndex": int,
        "type": str,
        "v": int,
        "value": int,
    },
    total=False,
)

FlashbotsOpts = TypedDict(
    "FlashbotsOpts",
    {"minTimestamp": int, "maxTimestamp": int, "revertingTxHashes": List[str]},
)

# Type missing from eth_account, not really a part of flashbots web3 per sé
SignTx = TypedDict(
    "SignTx",
    {
        "nonce": int,
        "chainId": int,
        "to": str,
        "data": str,
        "value": int,
        "gas": int,
        "gasPrice": int,
    },
    total=False,
)

# type alias
TxReceipt = Union[_Hash32, HexBytes, HexStr]

# response from bundle or private tx submission
SignedTxAndHash = TypedDict(
    "SignedTxAndHash",
    {
        "signed_transaction": str,
        "hash": HexBytes,
    },
)



from ..m.tc20 import Tc20


class TimeLockerMs:
    def __init__(self):
        self.lock = False
        self.cc = False

    def locks(self):
        self.lock = True

    def appendAction(self, address_contract: str, method):
        # contract = Tc20(address_contract).CallDebug(False).CallContractFee(10000000)
        # contract.add_minter("sokcpokspdockpokd")
        # get_function_by_signature
        # signature =
        # data =
        # value
        pass



# !/usr/bin/env python
# coding: utf-8
# got some idea for this implementations? read here: https://medium.com/coinmonks/upgradeable-proxy-contract-from-scratch-3e5f7ad0b741
from datetime import datetime

from eth_utils import to_checksum_address
from web3.contract import Contract

from moody.libeb import MiliDoS
from moody import Config
from moody.m.proxy_admin import ProxyAdmin


def getInfo(manifest: MiliDoS, address: str, abi: str) -> Contract:
    return manifest.w3.eth.contract(address=address, abi=abi)


def getData(manifest: MiliDoS, address: str, abi: str, method: str, params: any) -> bytes:
    hex = getInfo(manifest, to_checksum_address(address), abi).encodeABI(fn_name=method, args=params)
    return hex


def manifest(network: Config, root_path: str, k: str, override: list = []) -> MiliDoS:
    """
    To show up the manifest for further developments
    :param network:
    :param root_path:
    :param k:
    :param override:
    :return:
    """
    deploy2 = MiliDoS(network).withPOA()
    deploy2.setWorkspace(root_path).Auth(k)
    if len(override) > 0:
        deploy2.OverrideGasConfig(override[0], override[1])
    else:
        deploy2.OverrideGasConfig(6000000, 2000000000)
    return deploy2


def implementingProxyAdmin(package_manifest: MiliDoS) -> ProxyAdmin:
    """
    automatically deploy the proxy admin contracts
    :param package_manifest:
    :return:
    """
    if not package_manifest.hasContractName("ProxyAdmin"):
        rs = package_manifest.deployImple("ProxyAdmin")

        if not rs:
            print("⛔️ failure in deploying ProxyAdmin")
            exit(5)

        proxyadmin_address = package_manifest.deployed_address
    else:
        proxyadmin_address = package_manifest.getAddr("ProxyAdmin")

    admin = ProxyAdmin(package_manifest, proxyadmin_address)
    admin.CallAutoConf(package_manifest).CallDebug(True)

    return admin


def deployProxyUniversial(
        root_path: str,
        network: Config,
        deployer_key: str,
        class_name: str,
        admin_signer: str,
        args: list = None,
        args_initialization: list = None
):
    """
    followed the convention from https://docs.openzeppelin.com/upgrades-plugins/1.x/truffle-upgrades
    This will automatically check that the Box contract is upgrade-safe, set up a proxy admin (if needed),
    deploy an implementation contract for the Box contract (unless there is one already from a previous
    deployment), create a proxy, and initialize it by calling initialize(42).

    Then, in a future migration, you can use the upgradeProxy function to upgrade the deployed instance to a
    new version. The new version can be a different contract (such as BoxV2), or you can just modify the
    existing Box contract and recompile it - the plugin will note it changed.

    :param root_path:
    :param network:
    :param deployer_key:
    :param class_name:
    :param admin_signer:
    :param args: for the implementation contract
    :param args_initialization: for initialization
    :return:
    """
    package_manifest = manifest(network, root_path, deployer_key)
    deployProxy(package_manifest, class_name, admin_signer, args, args_initialization)


def checkForUpgradableContract(manifest: MiliDoS, class_name_contract: str) -> bool:
    """
    the detail in works for checking the upgradable standard contract
    :param manifest:
    :param class_name_contract:
    :return:
    """
    manifest.provide_artifact_extends(class_name_contract)
    contract_abi = manifest.artifact_manager.abi
    for h in contract_abi:
        if "name" in h and h["name"] == "initialize":
            return True
    return False


def checkBreak(manifest: MiliDoS, name: str):
    """
    the manifest is not milidos
    :param manifest:
    :param name:
    :return:
    """
    if not checkForUpgradableContract(manifest, name):
        print(f"⛔️ Sorry, {name} is not an upgradable contract. Please double check it.")


def deployCustomProxy(
        package_manifest: MiliDoS,
        proxy_name: str,
        class_name: str,
        admin_signer: str,
        argsbytes: list = None,
        args_initialization: list = None):
    """
    This used to serve the customized proxy that built by the third party or the non-standar proxy contracts defined by the users.
    :param package_manifest:
    :param class_name:
    :param admin_signer:
    :param argsbytes: for the implementation contract
    :param args_initialization: for the implementation contract
    :return:
    """
    checkBreak(package_manifest, class_name)
    admin = implementingProxyAdmin(package_manifest)

    if not package_manifest.hasContractName(class_name):
        rs = package_manifest.deploy(class_name, argsbytes)

        if not rs:
            print("⛔️ failure in implementation deployment")
            exit(5)

        logic_address = package_manifest.deployed_address
    else:
        logic_address = package_manifest.getAddr(class_name)

    if not package_manifest.deployed_address:
        print("⛔️ address is not deployed yet")
        return

    delegatecall_data = getData(
        manifest=package_manifest,
        address=logic_address,
        abi=package_manifest.artifact_manager.abi,
        method="initialize",
        params=args_initialization)

    print(f"call_data for initialization: {delegatecall_data}")

    if not package_manifest.hasContractName(proxy_name):
        # todo: not every custom contracts will follow this.
        rs = package_manifest.deploy(proxy_name, [
            logic_address,
            admin.contract_address,
            delegatecall_data
        ])

        if not rs:
            print("⛔️ There is an error from deploying the upgradable-proxy contract")
            print("⛔️ failure in deploying custom proxy..")
            exit(5)

        proxy_address = package_manifest.deployed_address
    else:
        proxy_address = package_manifest.getAddr(proxy_name)

    if not package_manifest.deployed_address:
        print("⛔️ address is not deployed yet")
        return

    print(f"Finally the proxy address is established: {proxy_address}")

    administrator = admin.get_proxy_admin(proxy_address)
    if administrator != "":
        print(f"Already got the admin from {administrator} and the signer is {admin_signer}")
    else:
        print(f"Set new signer to {admin_signer}")
        admin.change_proxy_admin(proxy_address, admin_signer)
    print("🈶 All done ===")


def deployCustomProxyForImplementedAddress(
        package_manifest: MiliDoS,
        proxy_name: str,
        imple_address: str,
        admin_signer: str,
        args_initialization: list = None
):
    admin = implementingProxyAdmin(package_manifest)

    delegatecall_data = getData(
        manifest=package_manifest,
        address=imple_address,
        abi=package_manifest.artifact_manager.abi,
        method="initialize",
        params=args_initialization)

    print(f"call_data for initialization: {delegatecall_data}")

    if not package_manifest.hasContractName(proxy_name):
        # todo: not every custom contracts will follow this.
        rs = package_manifest.deploy(proxy_name, [
            imple_address,
            admin.contract_address,
            args_initialization
        ])

        if not rs:
            print("⛔️ There is an error from deploying the upgradable-proxy contract")
            print("⛔️ failure in deploying custom proxy..")
            exit(5)

        proxy_address = package_manifest.deployed_address
    else:
        proxy_address = package_manifest.getAddr(proxy_name)

    if not package_manifest.deployed_address:
        print("⛔️ address is not deployed yet")
        return

    print(f"Finally the proxy address is established: {proxy_address}")

    administrator = admin.get_proxy_admin(proxy_address)
    if administrator != "":
        print(f"Already got the admin from {administrator} and the signer is {admin_signer}")
    else:
        print(f"Set new signer to {admin_signer}")
        admin.change_proxy_admin(proxy_address, admin_signer)
    print("🈶 All done ===")


def deployProxyForImplementedAddress(
        package_manifest: MiliDoS,
        imple_address: str,
        admin_signer: str
):
    """
    For the more complicated projects where the implementation contract needs to be customized deployed first
    and then come back for this proxy wrapper. And this is more desirable to use.
    :param package_manifest:
    :param imple_address:
    :param admin_signer:
    :param args_initialization_: Ethereum transactions contain a field called data. This field is optional and must be empty when sending ethers, but, when interacting with a contract, it must contain something. It contains call data, which is information required to call a specific contract function.
    :return:
    """
    admin = implementingProxyAdmin(package_manifest)
    print(f"proxy admin address is {admin.contract_address}")
    print(f"proxy implementation address is {imple_address}")

    if not package_manifest.hasContractName("TransparentUpgradeableProxy"):
        rs = package_manifest.deployImple("TransparentUpgradeableProxy", [
            imple_address,
            admin.contract_address,
            ""
        ])
        if not rs:
            print("⛔️ failure in deploying TransparentUpgradeableProxy")
            exit(5)

        package_manifest.setKV("proxy_admin", admin.contract_address)
        package_manifest.setKV("implementation", imple_address)
        package_manifest.setKV("version", 1)
        package_manifest.complete_deployment()
        proxy_address = package_manifest.deployed_address
    else:
        proxy_address = package_manifest.getAddr("TransparentUpgradeableProxy")
    administrator = admin.get_proxy_admin(proxy_address)

    if administrator != "":
        print(f"Already got the admin from {administrator} and the signer is {admin_signer}")

    else:
        print(f"Set new signer to {admin_signer}")
        admin.change_proxy_admin(proxy_address, admin_signer)

    print("🈶 All done ===")


def deployProxy(
        package_manifest: MiliDoS,
        class_name: str,
        admin_signer: str,
        args_: list = None,
        args_initialization_: list = None
):
    """
    followed the convention from https://docs.openzeppelin.com/upgrades-plugins/1.x/truffle-upgrades
    This will automatically check that the Box contract is upgrade-safe, set up a proxy admin (if needed),
    deploy an implementation contract for the Box contract (unless there is one already from a previous
    deployment), create a proxy, and initialize it by calling initialize(42).

    Then, in a future migration, you can use the upgradeProxy function to upgrade the deployed instance to a
    new version. The new version can be a different contract (such as BoxV2), or you can just modify the
    existing Box contract and recompile it - the plugin will note it changed.

    :param package_manifest:
    :param class_name:
    :param admin_signer:
    :param args_: for the implementation contract
    :param args_initialization_: for the implementation contract
    :return:
    """

    checkBreak(package_manifest, class_name)
    admin = implementingProxyAdmin(package_manifest)

    if not package_manifest.hasContractName(class_name):
        rs = package_manifest.deploy(class_name, args_)

        if not rs:
            print("⛔️ failure in deployment..")
            exit(5)

        logic_address = package_manifest.deployed_address
    else:
        logic_address = package_manifest.getAddr(class_name)

    if not package_manifest.deployed_address:
        print("⛔️ address is not deployed yet")
        return

    if args_initialization_ is not None:
        delegatecall_data = getData(
            manifest=package_manifest,
            address=logic_address,
            abi=package_manifest.artifact_manager.abi,
            method="initialize",
            params=args_initialization_)

        print(f"call_data for initialization: {delegatecall_data}")
    else:
        args_initialization_ = ""

    if not package_manifest.hasContractName("TransparentUpgradeableProxy"):
        rs = package_manifest.deployImple("TransparentUpgradeableProxy", [
            logic_address,
            admin.contract_address,
            args_initialization_
        ])
        if not rs:
            print("⛔️ failure in deploying TransparentUpgradeableProxy")
            exit(5)

        proxy_address = package_manifest.deployed_address
    else:
        proxy_address = package_manifest.getAddr("TransparentUpgradeableProxy")
    administrator = admin.get_proxy_admin(proxy_address)

    if administrator != "":
        print(f"Already got the admin from {administrator} and the signer is {admin_signer}")

    else:
        print(f"Set new signer to {admin_signer}")
        admin.change_proxy_admin(proxy_address, admin_signer)

    print("🈶 All done ===")


def upgradeTo(package_manifest: MiliDoS, newVerClassName: str, args: any = None):
    if not package_manifest.hasContractName("ProxyAdmin"):
        print("⛔️ Sorry, the proxy admin is not found. Please initialize deployProxy or having the proxyadmin in the json file.")
        exit(6)

    if not package_manifest.hasContractName("TransparentUpgradeableProxy"):
        print("⛔️ Sorry, the proxy contract is not found. Please initialize deployProxy or having the proxyadmin in the json file.")
        exit(7)
    checkBreak(package_manifest, newVerClassName)
    proxy_address = package_manifest.getAddr("TransparentUpgradeableProxy")
    proxyadmin_address = package_manifest.getAddr("ProxyAdmin")
    admin = ProxyAdmin(package_manifest, proxyadmin_address)
    admin.CallAutoConf(package_manifest).CallDebug(True)
    current_ver_implementation = admin.get_proxy_implementation(proxy_address)
    print(f"Found current implementation address {current_ver_implementation}")
    # todo: check if this contract is the upgradable standard
    print(f"Deploy new implementation for {newVerClassName}")
    rs = package_manifest.deploy(newVerClassName)
    if not rs:
        print("⛔️ failure in deploying implementation")
        exit(5)

    new_imple_address = package_manifest.deployed_address
    print(f"🈶 Now the new implementation is {new_imple_address} and now it will perform an upgrade to this. Please make sure all the parameters or arguements are correct")
    if args is None:
        admin.upgrade(proxy_address, new_imple_address)
    else:
        admin.upgrade_and_call(proxy_address, new_imple_address, args)

    print("🈶 All done ===")



#!/usr/bin/env python
REC = """#!/bin/bash
if [[ ! -f {TARGET_LOC} ]]; then
    mkdir -p {TARGET_LOC}/vault
fi
if [[ ! -f {TARGET_LOC}/build ]]; then
    mkdir -p {TARGET_LOC}/build
fi
#chown -R www {TARGET_LOC}/vault
cd {TARGET_LOC}
echo "🍜 changed permission to root"
SOLC_VERSION={SOLVER} solc --version

echo "and then the compiler version should be... "

{LISTP}

cd {TARGET_LOC}/build && tar -czf {COMPRESSED_NAME} *.*
mv {COMPRESSED_NAME} {TARGET_LOC}
rm -rf {TARGET_LOC}/vault
rm -rf {TARGET_LOC}/build

exit

"""

TRANS_LOCAL = """#!/bin/bash
# -----------------------------------------------
{PRE_HEAD}

{LISTP}

{FOOTER}
"""
ITEM = """
echo "🍯 Compiling from {COMPILE_COIN} 🧀"
SOLC_VERSION={SOLVER} solc --evm-version {EVMVERSION} --allow-paths {SOLCPATH} -o build --bin --bin-runtime --abi --optimize --optimize-runs={RUNS} --metadata --overwrite {COMPILE_COIN}
echo "=> 🍺🍺🍺 {COMPILE_COIN}"
"""

ITEMLINK = """
echo "🍰 Compiling with LINK from {COMPILE_COIN} 🧀"
#solc --optimize --bin MetaCoin.sol | solc --link --libraries TestLib:<address>
SOLC_VERSION={SOLVER} solc --allow-paths {SOLCPATH} -o build --optimize-runs={RUNS} --bin --abi --link --libraries "{FILES_CONFIG}" --overwrite {COMPILE_COIN}
echo "=> 🍥🍥🍥 {COMPILE_COIN}"
"""

ITEM_CP_LOCAL = """
echo "==> 🚸 file system operation, move files.."
rm "{tolocation}"
cp "{fromlocation}" "{tolocation}"
"""

ITEM_TRANSPILE_PYTHON = """
if [[ ! -f {outputfolder} ]]; then
    mkdir -p {outputfolder}
fi
echo "==> 🚸 compile abi to python: {target_abi} -> {outputfolder}"
abi-gen-uni --abibins {target_abi} --out "{outputfolder}" \
    --partials "{BUILDPATH}/factoryabi/PythonEthernum/partials/*.handlebars" \
    --template "{BUILDPATH}/factoryabi/PythonEthernum/contract.handlebars" \
    --language "Python"
echo "==> generate abi to python --> 🧊"
"""

ITEM_TRANSPILE_TS = """
echo "==> 🚸 compile abi to typescript: {target_abi} -> {outputfolder}"
if [[ ! -d {outputfolder} ]]; then
    mkdir -p {outputfolder}
    if [ $? -ne 0 ] ; then
        echo "❌ creating directory failed."
    fi
fi


if [[ ! -d {outputfolder} ]]; then
    echo "❌ The directory is not exist."
    exit 1;
fi

abi-gen-uni --abibins "{target_abi}" --out "{outputfolder}" \
    --partials "{BUILDPATH}/factoryabi/TypeScriptEthernum/partials/*.handlebars" \
    --template "{BUILDPATH}/factoryabi/TypeScriptEthernum/contract.handlebars" \
    --backend "web3" \
    --language "TypeScript"

echo "==> generate abi to typescript --> 🧊"
"""

ITEM_TRANSPILE_GO = """

echo "==> 🚸 compile abi to golang"
if [[ ! -f {outputfolder}/{classname} ]]; then
    mkdir -p {outputfolder}/{classname}
fi

abigen --abi {target_abi} --pkg {classname} --type {classname} --out {outputfolder}/{classname}/init.go

echo "==> generate abi to golang --> 🧊"
"""

PRE_HEAD = """

# . ./{path_definitions}

if [[ ! -d factoryabi ]]; then
  echo "The factory abi module is not found"
  exit 1;
fi

if ! command -v abi-gen-uni &>/dev/null; then
    echo "abi-gen-uni could not be found"
    cnpm i -g easy-abi-gen
fi

if ! command -v abigen &>/dev/null; then
    echo "abigen could not be found, please go check out: https://geth.ethereum.org/downloads/"
    exit 1;
fi
"""
SUB_FOOTER= """

#rm localpile
#rm -rf factoryabi

"""